PROGRAM_NAME='PLENA PLM-4Px2x Demo'
(***********************************************************)
(* System Type : NetLinx                                   *)
(***********************************************************)
(*  		    DEMO PROGRAM FOR                       *) 
(*       BOSCH PLENA PLM-4Px2x DSP POWER AMPIFIERS         *)
(*                                                         *)
(*      	BOSCH COMMUNICATIONS SYSTEMS               *)
(*    		    25 HUNTINGWOOD DRIVE                   *)
(*    		  HUNTINGWOOD NSW AUSTRALIA                *)
(*                                                         *)
(*	            VERSION 1.0 AUG 2013                   *)
(***********************************************************)
(*    DEVELOPMENT SYSTEM :-			           *)
(*      NI-3000 	F/W v2.32.148		           *)
(*	NXD-500I 	F/W v2.4.5			   *)
(***********************************************************)
(*          DEVICE NUMBER DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_DEVICE
dvTPA				= 10001:1:1	// TOUCH PANEL - GENERAL
dvTPB				= 10001:2:1	// TOUCH PANEL - AMP 1
dvTPC				= 10001:3:1	// TOUCH PANEL - AMP 2


vdCOMMS				= 33001:1:0	// COMMS DEVICE
vdAMP1				= 33002:1:0	// AMP1
vdAMP2				= 33003:1:0	// AMP2

(***********************************************************)
(*               CONSTANT DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_CONSTANT

INTEGER		MAX_ENTRIES	= 10	
INTEGER    	MAX_ENTRY_SIZE	= 128

CHAR 		FADER_BTNS[] 	= {11, 21, 31, 41}
DEV		TP_AMPS[]	= {dvTPB, dvTPC}	// AMPLIFIER TPS
(***********************************************************)
(*              DATA TYPE DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_TYPE

(***********************************************************)
(*               VARIABLE DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_VARIABLE

CHAR		AMP1_ADDR[]	= '192.168.0.31'
CHAR		AMP2_ADDR[]	= '192.168.0.32'

INTEGER		BASE_PORT 	= 10
INTEGER		AMP1_ID		= 1
INTEGER		AMP2_ID		= 2

INTEGER		CURRENT_FADER 	= 0			// CURRENT FADER BEING PRESSED
CHAR		PAGE_NAME[20] 	= 'Title'		// CURRENT PAGE NAME
CHAR		TERMINAL_ONLINE = 0

CHAR		ENTRY[MAX_ENTRIES][MAX_ENTRY_SIZE]	// DECODED DATA ENTRIES
INTEGER		NO_ENTRIES				// NUMBER OF ENTRIES

INTEGER		AMP1_LEVEL, AMP1_VALUE

CHAR		TAPER_TABLE[256], 
		ANTI_TAPER_TABLE[256]
(***********************************************************)
(*               LATCHING DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_LATCHING

(***********************************************************)
(*       MUTUALLY EXCLUSIVE DEFINITIONS GO BELOW           *)
(***********************************************************)
DEFINE_MUTUALLY_EXCLUSIVE

(***********************************************************)
(*        SUBROUTINE/FUNCTION DEFINITIONS GO BELOW         *)
(***********************************************************)
DEFINE_FUNCTION TERMINAL (CHAR S[])
{
//    SEND_STRING 0, S
}

DEFINE_FUNCTION CHAR[255] HEX_STRING (S_[])
{
    INTEGER X
    CHAR RES[255]
  
    RES = ''
    FOR (X = 1; X <= LENGTH_STRING (S_); X ++)
	RES = "RES, ' ', FORMAT ('%.2X', S_[X])"
    RETURN RES
}  

DEFINE_FUNCTION CHAR[255] DISPLAY_LINE (CHAR S_[])  
{
    INTEGER X
    CHAR RES[255]
  
    RES = ''
    FOR (X = 1; X <= LENGTH_STRING (S_); X ++)
	IF (S_[X] >= ' ' && S_[X] <= '~')
	    RES = "RES, S_[X]"
	ELSE  
	    RES = "RES, '<', ITOA (S_[X]), '>'"
    RETURN RES
}

DEFINE_FUNCTION CHAR[12] LVL_TO_DB (INTEGER LEVEL_)
{
    CHAR TMP[12]
    SINTEGER X

    IF (LEVEL_ >= 249)
	RETURN '24dB'
    ELSE IF (LEVEL_ <= 1)
	RETURN '-100.0dB'
    ELSE
    {
	X = (TYPE_CAST (LEVEL_) - 201) * 5
	RETURN "ITOA (X / 10), '.', ITOA (ABS_VALUE (X % 10)), 'dB'"
    }
}

DEFINE_FUNCTION SHOW_FEEDBACK ()
{
    [dvTPA, 10] = (PAGE_NAME = 'Master')
    [dvTPA, 11] = (PAGE_NAME = 'Channel 1')
    [dvTPA, 12] = (PAGE_NAME = 'Channel 2')
    [dvTPA, 13] = (PAGE_NAME = 'Channel 3')
    [dvTPA, 14] = (PAGE_NAME = 'Channel 4')    
    [dvTPA, 15] = (PAGE_NAME = 'Presets')    
}

DEFINE_FUNCTION SHOW_PAGE (CHAR PAGE_[])
{
    PAGE_NAME = PAGE_
    SEND_COMMAND dvTPA, "'PAGE-', PAGE_"
    SHOW_FEEDBACK ()
}

DEFINE_FUNCTION DECODE_DATA (INTEGER NO_, CHAR DATA_[])
{
    INTEGER I, X, Y, Z
	
    NO_ENTRIES = 0
    X = FIND_STRING (DATA_, ':', 1)
    WHILE (X)
    {
	NO_ENTRIES ++
	ENTRY[NO_ENTRIES] = LEFT_STRING (DATA_, X - 1)
	GET_BUFFER_STRING (DATA_, X)
	X = FIND_STRING (DATA_, ':', 1)
    }
    NO_ENTRIES ++
    ENTRY[NO_ENTRIES] = DATA.TEXT
//    FOR (X = 1; X <= NO_ENTRIES; X ++)
//	TERMINAL ("'  ', ITOA (X), '  "', ENTRY[X], '"'")
    IF (NO_ENTRIES = 1)
    {
	SWITCH (ENTRY[1])
	{
	    CASE 'ONLINE'  : TERMINAL ("'AMP ', ITOA (NO_), ' ONLINE'")
	    CASE 'OFFLINE' : TERMINAL ("'AMP ', ITOA (NO_), ' OFFLINE'")
	}
    }
    ELSE IF (NO_ENTRIES = 2)
    {
	SWITCH (ENTRY[1])
	{
	    CASE 'NAME'     : SEND_COMMAND TP_AMPS[NO_], "'^TXT-9,1&2,', ENTRY[2]"
	    CASE 'FAULTS'   : // ENTRY[2] HAS BITMAPPED FAULTS
	    {
		X = ATOI (ENTRY[2])
		[TP_AMPS[NO_], 61] = (X & $01 > 0)
		[TP_AMPS[NO_], 62] = (X & $02 > 0)
		[TP_AMPS[NO_], 63] = (X & $04 > 0)
		[TP_AMPS[NO_], 64] = (X & $08 > 0)
		[TP_AMPS[NO_], 65] = (X & $10 > 0)
		[TP_AMPS[NO_], 66] = (X & $20 > 0)
		[TP_AMPS[NO_], 67] = (X & $40 > 0)
		[TP_AMPS[NO_], 68] = (X & $80 > 0)
	    }
	    CASE 'POWER'    : [TP_AMPS[NO_], 2] = (ENTRY[2] = '1')
	    CASE 'MUTE'     : [TP_AMPS[NO_], 4] = (ENTRY[2] = '1')
	    CASE 'OVERRIDE' : [TP_AMPS[NO_], 6] = (ENTRY[2] = '1')
	    CASE 'ONLINE'   : [TP_AMPS[NO_], 8] = (ENTRY[2] = '1')
	    CASE 'ERR'      : TERMINAL ("'AMP ', ITOA (NO_), ' ERROR $', ENTRY[2]") 
	}
    }
    ELSE IF (NO_ENTRIES = 3)
    {
	X = ATOI (ENTRY[2])
	SWITCH (ENTRY[1])
	{
	    CASE 'INPUT_NAME'  : SEND_COMMAND TP_AMPS[NO_], "'^TXT-', ITOA (X), ',1&2,', ENTRY[3]"
	    CASE 'OUTPUT_NAME' : SEND_COMMAND TP_AMPS[NO_], "'^TXT-', ITOA (X + 4), ',1&2,', ENTRY[3]"
	}
    }
    ELSE IF (NO_ENTRIES = 4)
    {
	SWITCH (ENTRY[1])
	{
	    CASE 'OUTPUT' :
	    {
		X = ATOI (ENTRY[2])
		Y = ATOI (ENTRY[3])
		TERMINAL ("'OUTPUT ', ITOA (X), '  ', LVL_TO_DB (Y), ' MUTE ', ENTRY[4]")
		[TP_AMPS[NO_], X + 10] = (ENTRY[4] = '1')
		IF (CURRENT_FADER <> X) SEND_LEVEL TP_AMPS[NO_], X, ANTI_TAPER_TABLE[Y + 1]
		SEND_COMMAND TP_AMPS[NO_], "'^TXT-', ITOA (X + 20), ',1&2,', LVL_TO_DB (Y)"
	    }
	    CASE 'PRESET'      : 
	    {
		X = ATOI (ENTRY[2])
		Y = ATOI (ENTRY[3])
		SEND_COMMAND TP_AMPS[NO_], "'^TXT-', ITOA (X + 10), ',1&2,', ENTRY[4]"
		[TP_AMPS[NO_], X + 70] = (Y)
	    }
	}
    }
    ELSE IF (NO_ENTRIES = 5)
    {
	SWITCH (ENTRY[1])
	{
	    CASE 'MIX' :
	    {
		X = (ATOI (ENTRY[2]) - 1) * 4 + ATOI (ENTRY[3])
		Y = ATOI (ENTRY[4])
		TERMINAL ("'MIX ', ITOA (X), '  ',  LVL_TO_DB (Y), ' MUTE ', ENTRY[5]")
		[TP_AMPS[NO_], X + 20] = (ENTRY[5] = '1')
		IF (CURRENT_FADER <> X + 4) SEND_LEVEL TP_AMPS[NO_], X + 4, ANTI_TAPER_TABLE[Y + 1]
		SEND_COMMAND TP_AMPS[NO_], "'^TXT-', ITOA (X + 24), ',1&2,', LVL_TO_DB (Y)"
	    }
	}
    }
}

DEFINE_FUNCTION INIT ()
{
    INTEGER I
    
    FOR (I = 1; I <= MAX_ENTRIES; I ++)
	ENTRY[I] = ''
    NO_ENTRIES = 0
   
    TAPER_TABLE =	// THESE TABLES CONVERT THE FADERS FROM LIN-LOG TO LOG-LOG TAPER  
	"$00, $03, $05, $08, $0B, $0D, $10, $12, $15, $18, $1A, $1D, $1F, $22, $25, $27,
	 $2A, $2C, $2F, $31, $34, $36, $39, $3B, $3E, $40, $43, $45, $48, $4A, $4C, $4F,
	 $51, $53, $56, $58, $5A, $5C, $5F, $61, $63, $65, $67, $69, $6B, $6D, $6F, $71,
	 $73, $75, $77, $79, $7B, $7C, $7E, $80, $81, $83, $85, $86, $88, $89, $8A, $8C,
	 $8D, $8E, $8F, $91, $92, $93, $94, $95, $96, $97, $97, $98, $99, $9A, $9B, $9B,
	 $9C, $9D, $9D, $9E, $9E, $9F, $A0, $A0, $A1, $A1, $A2, $A2, $A3, $A3, $A4, $A4,
	 $A5, $A6, $A6, $A7, $A7, $A8, $A8, $A9, $A9, $AA, $AA, $AB, $AB, $AC, $AC, $AD,
	 $AD, $AE, $AE, $AF, $AF, $B0, $B0, $B1, $B1, $B2, $B2, $B3, $B3, $B4, $B4, $B5,
	 $B5, $B5, $B6, $B6, $B7, $B7, $B7, $B8, $B8, $B8, $B9, $B9, $B9, $BA, $BA, $BA,
	 $BA, $BB, $BB, $BB, $BC, $BC, $BC, $BC, $BD, $BD, $BD, $BE, $BE, $BE, $BE, $BF,
	 $BF, $BF, $C0, $C0, $C0, $C0, $C1, $C1, $C1, $C2, $C2, $C2, $C3, $C3, $C3, $C4,
	 $C4, $C4, $C5, $C5, $C5, $C5, $C6, $C6, $C6, $C7, $C7, $C7, $C8, $C8, $C8, $C9,
	 $C9, $C9, $CA, $CA, $CA, $CB, $CB, $CB, $CC, $CC, $CC, $CC, $CD, $CD, $CD, $CE,
	 $CE, $CE, $CF, $CF, $CF, $D0, $D0, $D0, $D1, $D1, $D1, $D1, $D2, $D2, $D2, $D3,
	 $D3, $D3, $D4, $D4, $D4, $D5, $D5, $D5, $D5, $D6, $D6, $D6, $D7, $D7, $D7, $D8,
	 $D8, $D8, $D9, $D9, $D9, $DA, $DA, $DA, $DA, $DB, $DB, $DB, $DC, $DC, $DC, $DD" 

    ANTI_TAPER_TABLE = 
	"$00, $01, $01, $01, $02, $02, $03, $03, $03, $04, $04, $04, $05, $05, $06, $06,
	 $06, $07, $07, $08, $08, $08, $09, $09, $09, $0A, $0A, $0B, $0B, $0B, $0C, $0C,
	 $0D, $0D, $0D, $0E, $0E, $0E, $0F, $0F, $10, $10, $10, $11, $11, $12, $12, $12,
	 $13, $13, $14, $14, $14, $15, $15, $16, $16, $16, $17, $17, $18, $18, $18, $19,
	 $19, $1A, $1A, $1A, $1B, $1B, $1C, $1C, $1C, $1D, $1D, $1E, $1E, $1F, $1F, $1F,
	 $20, $20, $21, $21, $22, $22, $22, $23, $23, $24, $24, $25, $25, $26, $26, $26,
	 $27, $27, $28, $28, $29, $29, $2A, $2A, $2B, $2B, $2C, $2C, $2D, $2D, $2E, $2E,
	 $2F, $2F, $30, $30, $31, $31, $32, $32, $33, $33, $34, $34, $35, $36, $36, $37,
	 $37, $38, $39, $39, $3A, $3A, $3B, $3C, $3C, $3D, $3E, $3F, $3F, $40, $41, $42,
	 $43, $43, $44, $45, $46, $47, $48, $49, $4B, $4C, $4D, $4E, $50, $51, $53, $55,
	 $56, $58, $5A, $5C, $5E, $60, $61, $63, $65, $67, $69, $6B, $6D, $6F, $71, $73,
	 $75, $77, $79, $7B, $7D, $7F, $82, $84, $87, $8A, $8D, $91, $94, $98, $9B, $9F,
	 $A2, $A6, $A9, $AC, $AF, $B2, $B6, $B9, $BC, $BF, $C2, $C5, $C8, $CC, $CF, $D2,
	 $D5, $D8, $DC, $DF, $E2, $E5, $E9, $EC, $EF, $F2, $F5, $F9, $FC, $FF, $FF, $FF,
	 $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
	 $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF" 
}

(***********************************************************)
(*                STARTUP CODE GOES BELOW                  *)
(***********************************************************)
DEFINE_START
INIT ()
SET_VIRTUAL_LEVEL_COUNT (dvTPB, 20)	// FIX FOR NXD-500i
SET_VIRTUAL_LEVEL_COUNT (dvTPC, 20)	// FIX FOR NXD-500i

DEFINE_MODULE 'PLENA Comms' COMMS1 (vdCOMMS, BASE_PORT)
DEFINE_MODULE 'PLENA PLM-4Px2x' PLM1 (vdAMP1, vdCOMMS, AMP1_ADDR, AMP1_ID)
DEFINE_MODULE 'PLENA PLM-4Px2x' PLM2 (vdAMP2, vdCOMMS, AMP2_ADDR, AMP2_ID)

(***********************************************************)
(*                THE EVENTS GO BELOW                      *)
(***********************************************************)
DEFINE_EVENT

DATA_EVENT [dvTPA]
{
    ONLINE  : SHOW_PAGE ('Title')
}

DATA_EVENT [dvTPB]
{
    ONLINE : SEND_COMMAND vdAMP1, 'REFRESH'
}

DATA_EVENT [dvTPC]
{
    ONLINE : SEND_COMMAND vdAMP2, 'REFRESH'
}

BUTTON_EVENT [dvTPA, 0]
{
    PUSH    :
    {
	SWITCH (BUTTON.INPUT.CHANNEL)
	{
	    CASE 10 : SHOW_PAGE ('Master')
	    CASE 11 :
	    CASE 12 :
	    CASE 13 :
	    CASE 14 : SHOW_PAGE ("'Channel ', ITOA (BUTTON.INPUT.CHANNEL - 10)")
	    CASE 15 : SHOW_PAGE ('Presets')
	}
	SHOW_FEEDBACK ()
    }
    RELEASE :
    {
    }
}

BUTTON_EVENT [dvTPB, 0] 
{
    PUSH    : 
    {
	INTEGER BTN
	
	BTN = BUTTON.INPUT.CHANNEL
	TERMINAL ("'TPB PUSH ', ITOA (BTN)")
	SWITCH (BTN)
	{
	    CASE  2 : SEND_COMMAND vdAMP1, 'POWER:T'
	    CASE  4 : SEND_COMMAND vdAMP1, 'MUTE:T'
	
	    CASE 11 : SEND_COMMAND vdAMP1, 'OUTPUT_MUTE:1:T'
	    CASE 12 : SEND_COMMAND vdAMP1, 'OUTPUT_MUTE:2:T'
	    CASE 13 : SEND_COMMAND vdAMP1, 'OUTPUT_MUTE:3:T'
	    CASE 14 : SEND_COMMAND vdAMP1, 'OUTPUT_MUTE:4:T'
	    
	    CASE 21 : SEND_COMMAND vdAMP1, 'MIX_MUTE:1:1:T'
	    CASE 22 : SEND_COMMAND vdAMP1, 'MIX_MUTE:1:2:T'
	    CASE 23 : SEND_COMMAND vdAMP1, 'MIX_MUTE:1:3:T'
	    CASE 24 : SEND_COMMAND vdAMP1, 'MIX_MUTE:1:4:T'
	    
	    CASE 25 : SEND_COMMAND vdAMP1, 'MIX_MUTE:2:1:T'
	    CASE 26 : SEND_COMMAND vdAMP1, 'MIX_MUTE:2:2:T'
	    CASE 27 : SEND_COMMAND vdAMP1, 'MIX_MUTE:2:3:T'
	    CASE 28 : SEND_COMMAND vdAMP1, 'MIX_MUTE:2:4:T'
	    
	    CASE 29 : SEND_COMMAND vdAMP1, 'MIX_MUTE:3:1:T'
	    CASE 30 : SEND_COMMAND vdAMP1, 'MIX_MUTE:3:2:T'
	    CASE 31 : SEND_COMMAND vdAMP1, 'MIX_MUTE:3:3:T'
	    CASE 32 : SEND_COMMAND vdAMP1, 'MIX_MUTE:3:4:T'
	    
	    CASE 33 : SEND_COMMAND vdAMP1, 'MIX_MUTE:4:1:T'
	    CASE 34 : SEND_COMMAND vdAMP1, 'MIX_MUTE:4:2:T'
	    CASE 35 : SEND_COMMAND vdAMP1, 'MIX_MUTE:4:3:T'
	    CASE 36 : SEND_COMMAND vdAMP1, 'MIX_MUTE:4:4:T'
	    
	    CASE 81 : SEND_COMMAND vdAMP1, 'PRESET:1'
	    CASE 82 : SEND_COMMAND vdAMP1, 'PRESET:2'
	    CASE 83 : SEND_COMMAND vdAMP1, 'PRESET:3'
	    CASE 84 : SEND_COMMAND vdAMP1, 'PRESET:4'
	    CASE 85 : SEND_COMMAND vdAMP1, 'PRESET:5'
	    
	    DEFAULT :
	    {
		IF (BTN > 100 && BTN <= 120)
		    CURRENT_FADER = BTN - 100
	    }
	}
    }
    RELEASE : 
    {
	INTEGER BTN
	
	BTN = BUTTON.INPUT.CHANNEL
	TERMINAL ("'TPB RELEASE ', ITOA (BTN)")
	SWITCH (BTN)
	{
	    CASE  1 : ;
	    DEFAULT :
	    {
		IF (BTN > 100 && BTN <= 120)
		    CURRENT_FADER = 0
	    }
	}
    }
}

LEVEL_EVENT [dvTPB, 0]
{
    INTEGER X, Y
    
    IF (CURRENT_FADER = LEVEL.INPUT.LEVEL) // LEVEL BEING CHANGED ON TP
    {
	TERMINAL ("'TPB ', ITOA (LEVEL.INPUT.LEVEL), ' = ', ITOA (LEVEL.VALUE), ' (', ITOA (TAPER_TABLE[LEVEL.VALUE + 1]), ')'")
	IF (LEVEL.INPUT.LEVEL <= 4)
	    SEND_COMMAND vdAMP1, "'OUTPUT:', ITOA (LEVEL.INPUT.LEVEL), ':', ITOA (TAPER_TABLE[LEVEL.VALUE + 1])"
	ELSE IF (LEVEL.INPUT.LEVEL <= 20)
	{
	    X = (LEVEL.INPUT.LEVEL - 1) / 4
	    Y = LEVEL.INPUT.LEVEL % 4
	    IF (Y = 0) Y = 4
	    SEND_COMMAND vdAMP1, "'MIX:', ITOA (X), ':', ITOA (Y), ':', ITOA (TAPER_TABLE[LEVEL.VALUE + 1])"
	}
    }
}

// DECODE STRING RESPONSES FROM AMPLIFIER MODULES
DATA_EVENT [vdAMP1]
{
    STRING : DECODE_DATA (1, DATA.TEXT)
}

DATA_EVENT [vdAMP2]
{
    STRING : DECODE_DATA (2, DATA.TEXT)
}

(***********************************************************)
(*            THE ACTUAL PROGRAM GOES BELOW                *)
(***********************************************************)
DEFINE_PROGRAM

(***********************************************************)
(*                     END OF PROGRAM                      *)
(*        DO NOT PUT ANY CODE BELOW THIS COMMENT           *)
(***********************************************************)

