MODULE_NAME='PLENA PLM-4Px2x' (DEV VD_AMP,	// VIRTUAL DEVICE OF AMP 
			       DEV VD_COMMS, 	// VIRTUAL DEVICE OF CENTRAL COMMS
			       CHAR AMP_ADDR[],	// AMP ADDRESS
			       INTEGER AMP_NO)  // AMP NUMBER

(***********************************************************)
(* System Type : NetLinx                                   *)
(***********************************************************)
(*             AMPLIFIER INTERFACE MODULE FOR              *) 
(*       BOSCH PLENA PLM-4Px2x DSP POWER AMPIFIERS         *)
(*                                                         *)
(*      	BOSCH COMMUNICATIONS SYSTEMS               *)
(*    		    25 HUNTINGWOOD DRIVE                   *)
(*    		  HUNTINGWOOD NSW AUSTRALIA                *)
(*                                                         *)
(*	            VERSION 1.0 AUG 2013                   *)
(***********************************************************)
(*
    NO.		BUTTON		CHANNEL

    1		CONNECT		CONNECTED
    2		DISCONNECT	DISCONNECTED

    11		CHANNEL 1 MUTE	CHANNEL 1 MUTED
    12		CHANNEL 2 MUTE	CHANNEL 2 MUTED
    13		CHANNEL 3 MUTE	CHANNEL 3 MUTED
    14		CHANNEL 4 MUTE	CHANNEL 4 MUTED

    21		CHANNEL 1 MUTE	CHANNEL 1 MUTED
    22		CHANNEL 2 MUTE	CHANNEL 2 MUTED
    23		CHANNEL 3 MUTE	CHANNEL 3 MUTED
    24		CHANNEL 4 MUTE	CHANNEL 4 MUTED

*)
(***********************************************************)
(*          DEVICE NUMBER DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_DEVICE

(***********************************************************)
(*               CONSTANT DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_CONSTANT

(* SYSTEM *)
NOS_LEVELS			= 4
MAX_AMPS			= 10
MAX_ENTRIES			= 10
MAX_ENTRY_SIZE			= 128
QUEUE_SIZE			= 100

(* AMPLIFIER *)
TX_PORT				= 12128		// TRASMISSION UDP PORT
RX_PORT				= 12129		// RECEPTION UDP PORT

MIXER_ID			= $5E40
AMP_ID				= $5E41

MASTER_ID			= $0001
SLAVE_ID			= $0100

dsHEADER			= 1
dsCMD				= 2
dsDATA				= 3

lsOFFLINE			= 0
lsLOOKING			= 1
lsFOUND				= 2

(* GLOABL OBJECTS *)
GOBJ_FORCE_STANDBY		= 1
GOBJ_GLOBAL_MUTE_ALLOWED	= 15
GOBJ_FORCE_GLOBAL_MUTE		= 16

(* PRESET OBJECTS *)
POBJ_AMPCH1_DSP_BASSENHANCE 	= 1 	// Amp Channel 1 Bass Enhance
POBJ_AMPCH1_CB_MIXINPUT1	= 20 	// Amp Channel 1 Mix Input 1 Level
POBJ_AMPCH1_CB_MIXINPUT2	= 21	// Amp Channel 1 Mix Input 2 Level
POBJ_AMPCH1_CB_MIXINPUT3	= 22	// Amp Channel 1 Mix Input 3 Level
POBJ_AMPCH1_CB_MIXINPUT4	= 23	// Amp Channel 1 Mix Input 4 Level
POBJ_AMPCH1_CB_OUTPUTLEVEL	= 26	// Amp Channel 1 Main Output Level

POBJ_AMPCH2_DSP_BASSENHANCE 	= 27 	// Amp Channel 2 Bass Enhance
POBJ_AMPCH2_CB_MIXINPUT1	= 46 	// Amp Channel 2 Mix Input 1 Level
POBJ_AMPCH2_CB_MIXINPUT2	= 47	// Amp Channel 2 Mix Input 2 Level
POBJ_AMPCH2_CB_MIXINPUT3	= 48	// Amp Channel 2 Mix Input 3 Level
POBJ_AMPCH2_CB_MIXINPUT4	= 49	// Amp Channel 2 Mix Input 4 Level
POBJ_AMPCH2_CB_OUTPUTLEVEL	= 52	// Amp Channel 2 Main Output Level

POBJ_AMPCH3_DSP_BASSENHANCE 	= 53 	// Amp Channel 3 Bass Enhance
POBJ_AMPCH3_CB_MIXINPUT1	= 72 	// Amp Channel 3 Mix Input 1 Level
POBJ_AMPCH3_CB_MIXINPUT2	= 73	// Amp Channel 3 Mix Input 2 Level
POBJ_AMPCH3_CB_MIXINPUT3	= 74	// Amp Channel 3 Mix Input 3 Level
POBJ_AMPCH3_CB_MIXINPUT4	= 75	// Amp Channel 3 Mix Input 4 Level
POBJ_AMPCH3_CB_OUTPUTLEVEL	= 78	// Amp Channel 3 Main Output Level
		
POBJ_AMPCH4_DSP_BASSENHANCE 	= 79 	// Amp Channel 4 Bass Enhance
POBJ_AMPCH4_CB_MIXINPUT1	= 98 	// Amp Channel 4 Mix Input 1 Level
POBJ_AMPCH4_CB_MIXINPUT2	= 99	// Amp Channel 4 Mix Input 2 Level
POBJ_AMPCH4_CB_MIXINPUT3	= 100	// Amp Channel 4 Mix Input 3 Level
POBJ_AMPCH4_CB_MIXINPUT4	= 101	// Amp Channel 4 Mix Input 4 Level
POBJ_AMPCH4_CB_OUTPUTLEVEL	= 104	// Amp Channel 4 Main Output Level

(* AMP COMMANDS *)
acPOWER				= 1	// Amp On / Standby
acCAN_MUTE			= 2	// Can mute Amp
acMUTE				= 3	// Amp All Mute
acPING				= 4	// Ping Amp
acSYNC				= 5	// Sync to Amp
acPASS				= 6	// Get Amp's password
acPRESET			= 7	// Recall Preset

acCH1IP1			= 11	// Amp Channel 1 Input 1 Level
acCH1IP2			= 12	// Amp Channel 1 Input 2 Level
acCH1IP3			= 13	// Amp Channel 1 Input 3 Level
acCH1IP4			= 14	// Amp Channel 1 Input 4 Level
acCH1				= 15	// Amp Channel 1 Output Level

acCH2IP1			= 21	// Amp Channel 2 Input 1 Level
acCH2IP2			= 22	// Amp Channel 2 Input 2 Level
acCH2IP3			= 23	// Amp Channel 2 Input 3 Level
acCH2IP4			= 24	// Amp Channel 2 Input 4 Level
acCH2				= 25	// Amp Channel 2 Output Level

acCH3IP1			= 31	// Amp Channel 3 Input 1 Level
acCH3IP2			= 32	// Amp Channel 3 Input 2 Level
acCH3IP3			= 33	// Amp Channel 3 Input 3 Level
acCH3IP4			= 34	// Amp Channel 3 Input 4 Level
acCH3				= 35	// Amp Channel 3 Output Level

acCH4IP1			= 41	// Amp Channel 4 Input 1 Level
acCH4IP2			= 42	// Amp Channel 4 Input 2 Level
acCH4IP3			= 43	// Amp Channel 4 Input 3 Level
acCH4IP4			= 44	// Amp Channel 4 Input 4 Level
acCH4				= 45	// Amp Channel 4 Output Level

(* TIME LINES *)
TL_TICK				= 1	// 100MS TICK
(***********************************************************)
(*              DATA TYPE DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_TYPE

STRUCTURE _AMP
{
    INTEGER     STATUS
    INTEGER     POLL
    INTEGER     TIMEOUT
    INTEGER 	POWERED
    INTEGER 	GLOBAL_MUTED	// GLOBAL MUTE
    INTEGER	OVERRIDEN
    INTEGER	FORCED_STANDBY
    INTEGER	FORCED_STANDBY_ALLOWED
    INTEGER 	MIX[16]
    INTEGER	MIXMUTED[16]
    INTEGER 	LEVELS[NOS_LEVELS]
    INTEGER	MUTED[NOS_LEVELS]
    CHAR	NAME[81]
    CHAR	INPUT_NAME[4][31]
    CHAR	OUTPUT_NAME[4][31]
    CHAR	PRESET_NAME[5][31]
    INTEGER	PRESET_VALID[5]
    CHAR	FAULTS
}

STRUCTURE _QUEUE	// QUEUE USED TO REGULATE LEVEL CHANGES
{
    INTEGER 	ID[QUEUE_SIZE]
    INTEGER 	PARAM[QUEUE_SIZE]
    INTEGER	IPTR, OPTR
    INTEGER 	COUNT
    INTEGER 	FULL
    INTEGER 	EMPTY
}

(***********************************************************)
(*               VARIABLE DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_VARIABLE

_AMP		AMP
_QUEUE		QUEUE
INTEGER		NO_ENTRIES
CHAR		ENTRY[MAX_ENTRIES][MAX_ENTRY_SIZE]
INTEGER		SEQ_NO
LONG		TL_TICK_COUNT[] = { 100 }

INTEGER		POLL_TIME = 60	// DEFAULT IS 6 SECONDS
INTEGER		PING_TIME = 120	// DEFAULT IS 12 SECONDS
INTEGER		RX_NO
CHAR		RX_CMD[4]
CHAR		RX_DATA[512]

(***********************************************************)
(*               LATCHING DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_LATCHING

(***********************************************************)
(*       MUTUALLY EXCLUSIVE DEFINITIONS GO BELOW           *)
(***********************************************************)
DEFINE_MUTUALLY_EXCLUSIVE

(***********************************************************)
(*        SUBROUTINE/FUNCTION DEFINITIONS GO BELOW         *)
(***********************************************************)
DEFINE_FUNCTION CHAR[255] DISPLAY_STRING (CHAR S[])
{
    CHAR RES[255]
    INTEGER I
 
    RES = ''
    FOR (I = 1; I <= LENGTH_STRING (S); I ++)
    {
	IF (S[I] >= ' ' && S[I] <= '~')
	    RES = "RES, S[I]"
	ELSE
	    RES = "RES, '<', ITOA (S[I]), '>'"
    }
  	RETURN RES
}

DEFINE_FUNCTION CHAR[255] HEX_STRING (S[])
{
    INTEGER X
    CHAR RES[255]
  
    RES = ''
    FOR (X = 1; X <= LENGTH_STRING (S); X ++)
	RES = "RES, ' ', FORMAT ('%.2X', S[X])"
    RETURN RES
}  

DEFINE_FUNCTION TERMINAL (CHAR S[])
{
//    SEND_STRING 0, S
}

DEFINE_FUNCTION CHAR[2] WORD (INTEGER W)
{
    RETURN "W / $100, W % $100"
}

DEFINE_FUNCTION INTEGER PRESENT (INTEGER ID_, INTEGER PARAM_)
{
    INTEGER I, RES
    
    RES = 0
    FOR (I = 1; I <= QUEUE_SIZE; I ++)
	IF (QUEUE.ID[I] = ID_ && (QUEUE.PARAM[I] = PARAM_ || PARAM_ = 0))
	    RES ++
    RETURN RES
}

DEFINE_FUNCTION UPDATE_NAMES ()
{
    INTEGER I
    
    SEND_STRING VD_AMP, "'NAME:', AMP.NAME"
    FOR (I = 1; I <= 4; I ++)
	SEND_STRING VD_AMP, "'INPUT_NAME:', ITOA (I), ':', AMP.INPUT_NAME[I]"
    FOR (I = 1; I <= 4; I ++)
	SEND_STRING VD_AMP, "'OUTPUT_NAME:', ITOA (I), ':', AMP.OUTPUT_NAME[I]"
}

DEFINE_FUNCTION UPDATE_PRESETS ()
{
    INTEGER I
    
    FOR (I = 1; I <= 5; I ++)
	SEND_STRING VD_AMP, "'PRESET:', ITOA (I), ':', ITOA (AMP.PRESET_VALID[I]),
			    ':', AMP.PRESET_NAME[I]"
}

DEFINE_FUNCTION UPDATE_STATUS ()
{
    SEND_STRING VD_AMP, "'FAULTS:', ITOA (AMP.FAULTS)"
    SEND_STRING VD_AMP, "'POWER:', ITOA (AMP.POWERED)"
    SEND_STRING VD_AMP, "'MUTE:', ITOA (AMP.GLOBAL_MUTED)"
    SEND_STRING VD_AMP, "'OVERRIDE:', ITOA (AMP.OVERRIDEN)"
    SEND_STRING VD_AMP, "'ONLINE:', ITOA (AMP.STATUS = lsFOUND)"
}

DEFINE_FUNCTION UPDATE_LEVELS ()
{
    INTEGER I, J, X
    
    FOR (I = 1; I <= 4; I ++)
    {
	SEND_STRING VD_AMP, "'OUTPUT:', ITOA (I), ':', ITOA (AMP.LEVELS[I]), ':', ITOA (AMP.MUTED[I])"
	FOR (J = 1; J <= 4; J ++)
	{
	    X = ((I - 1) * 4) + J;
	    SEND_STRING VD_AMP, "'MIX:', ITOA (I), ':', ITOA (J), ':', ITOA (AMP.MIX[X]), ':', ITOA (AMP.MIXMUTED[X])"
	}
    } 		
}

DEFINE_FUNCTION ADD (INTEGER ID_, INTEGER PARAM_)
{
    IF (!QUEUE.FULL)
    {       
	IF (QUEUE.IPTR = QUEUE_SIZE)
	    QUEUE.IPTR = 1;
    	ELSE
	    QUEUE.IPTR ++
	QUEUE.ID[QUEUE.IPTR] = ID_
    	QUEUE.PARAM[QUEUE.IPTR] = PARAM_
    	QUEUE.EMPTY = 0;
    	IF (QUEUE.IPTR = QUEUE_SIZE)
	    QUEUE.FULL = ((QUEUE.ID[1]) || (QUEUE.PARAM[1]));
    	ELSE
	    QUEUE.FULL = ((QUEUE.ID[QUEUE.IPTR + 1]) || 
			  (QUEUE.PARAM[QUEUE.IPTR + 1]));  
    } 
}

DEFINE_FUNCTION CHAR[255] WRITE_GLOBAL_OBJECT (INTEGER ID_, INTEGER COMMIT_, CHAR DATA_[])
{
    RETURN "$00, WORD (ID_), COMMIT_, DATA_"
}

DEFINE_FUNCTION CHAR[8] READ_GLOBAL_OBJECT (INTEGER ID_)
{
    RETURN "$01, WORD (ID_)"
}

DEFINE_FUNCTION CHAR[2] BOOL_BLOCK (INTEGER STATE_)
{
    IF (STATE_)
    	RETURN "$01, $3A"
    ELSE
	RETURN "$00, $64"
}

DEFINE_FUNCTION CHAR[255] WRITE_PRESET_OBJECT (INTEGER NO_, INTEGER ID_, CHAR DATA_[])
{
    RETURN "$00, NO_, WORD (ID_), $00, DATA_, $00"
}

DEFINE_FUNCTION CHAR[255] WRITE_AMP_LEVEL (INTEGER NO_, INTEGER ID_, INTEGER LEVEL_, INTEGER MUTE_)
{
    RETURN WRITE_PRESET_OBJECT (NO_, ID_, "LEVEL_, MUTE_")
}

DEFINE_FUNCTION SEND (CHAR CMD_[], CHAR DATA_[])
{
    SEND_COMMAND VD_COMMS, "AMP_NO, CMD_, DATA_"
    AMP.TIMEOUT =  60	// ALLOW 6 SECONDS TO COMPLETE
}

DEFINE_FUNCTION SEND_CMD (INTEGER ID_, INTEGER PARAM_)
{
    SWITCH (ID_)
    {
	CASE acPOWER    : SEND ('GOBJ', WRITE_GLOBAL_OBJECT (GOBJ_FORCE_STANDBY, 0, BOOL_BLOCK (PARAM_)))
	CASE acMUTE     : SEND ('GOBJ', WRITE_GLOBAL_OBJECT (GOBJ_FORCE_GLOBAL_MUTE, 0, BOOL_BLOCK (PARAM_)))
	CASE acCAN_MUTE : SEND ('GOBJ', WRITE_GLOBAL_OBJECT (GOBJ_GLOBAL_MUTE_ALLOWED, 0, BOOL_BLOCK (PARAM_)))
	CASE acPING	: SEND ('PING', '')
	CASE acPASS     : SEND ('PASS', '')
	CASE acSYNC	: SEND ('SYNC', "PARAM_")	
	CASE acPRESET   : SEND ('PSET', "PARAM_, PARAM_, $00")
	CASE acCH1      : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH1_CB_OUTPUTLEVEL, AMP.LEVELS[1], AMP.MUTED[1]))
	CASE acCH2      : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH2_CB_OUTPUTLEVEL, AMP.LEVELS[2], AMP.MUTED[2]))
	CASE acCH3      : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH3_CB_OUTPUTLEVEL, AMP.LEVELS[3], AMP.MUTED[3]))
	CASE acCH4      : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH4_CB_OUTPUTLEVEL, AMP.LEVELS[4], AMP.MUTED[4]))
	CASE acCH1IP1   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH1_CB_MIXINPUT1, AMP.MIX[1], AMP.MIXMUTED[1]))
	CASE acCH1IP2   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH1_CB_MIXINPUT2, AMP.MIX[2], AMP.MIXMUTED[2]))
	CASE acCH1IP3   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH1_CB_MIXINPUT3, AMP.MIX[3], AMP.MIXMUTED[3]))
	CASE acCH1IP4   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH1_CB_MIXINPUT4, AMP.MIX[4], AMP.MIXMUTED[4]))
	CASE acCH2IP1   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH2_CB_MIXINPUT1, AMP.MIX[5], AMP.MIXMUTED[5]))
	CASE acCH2IP2   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH2_CB_MIXINPUT2, AMP.MIX[6], AMP.MIXMUTED[6]))
	CASE acCH2IP3   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH2_CB_MIXINPUT3, AMP.MIX[7], AMP.MIXMUTED[7]))
	CASE acCH2IP4   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH2_CB_MIXINPUT4, AMP.MIX[8], AMP.MIXMUTED[8]))
	CASE acCH3IP1   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH3_CB_MIXINPUT1, AMP.MIX[9], AMP.MIXMUTED[9]))
	CASE acCH3IP2   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH3_CB_MIXINPUT2, AMP.MIX[10], AMP.MIXMUTED[10]))
	CASE acCH3IP3   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH3_CB_MIXINPUT3, AMP.MIX[11], AMP.MIXMUTED[11]))
	CASE acCH3IP4   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH3_CB_MIXINPUT4, AMP.MIX[12], AMP.MIXMUTED[12]))
	CASE acCH4IP1   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH4_CB_MIXINPUT1, AMP.MIX[13], AMP.MIXMUTED[13]))
	CASE acCH4IP2   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH4_CB_MIXINPUT2, AMP.MIX[14], AMP.MIXMUTED[14]))
	CASE acCH4IP3   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH4_CB_MIXINPUT3, AMP.MIX[15], AMP.MIXMUTED[15]))
	CASE acCH4IP4   : SEND ('POBJ', WRITE_AMP_LEVEL (0, POBJ_AMPCH4_CB_MIXINPUT4, AMP.MIX[16], AMP.MIXMUTED[16]))
    }
}

DEFINE_FUNCTION EXEC ()
{
    INTEGER CHK, X
    INTEGER ID_, PARAM_
    
    IF (!QUEUE.EMPTY)
    {
	IF (QUEUE.OPTR = QUEUE_SIZE)
	    QUEUE.OPTR = 1
	ELSE
	    QUEUE.OPTR ++
	ID_ = QUEUE.ID[QUEUE.OPTR]
	PARAM_ = QUEUE.PARAM[QUEUE.OPTR]
	QUEUE.ID[QUEUE.OPTR] = 0
	QUEUE.PARAM[QUEUE.OPTR] = 0
	QUEUE.FULL = 0
	QUEUE.EMPTY = 1
	SEND_CMD (ID_, PARAM_)
	QUEUE.COUNT = 1
	FOR (CHK = 1; CHK <= QUEUE_SIZE; CHK ++)
      	    IF (QUEUE.ID[CHK]) QUEUE.EMPTY = 0
    }
}       

DEFINE_FUNCTION DECODE_DATA ()
{
    INTEGER I, J, X, Y, LEN
    INTEGER L, M
    CHAR S[255]
    LONG RX_ERR_CODE

    AMP.TIMEOUT = 0
    LEN = LENGTH_STRING (RX_DATA)
    TERMINAL ("'RX ', ITOA (AMP_NO), ' CMD ', RX_CMD, ' DATA ', ITOA (LEN)")
    IF (RX_CMD = 'WHAT'  && LEN = 138)		// RESPONSE TO PING COMMAND
    {
    	S = MID_STRING (RX_DATA, 58, 81)
    	X = FIND_STRING (S, "$00", 1)
	IF (X) S = LEFT_STRING (S, X - 1)
	IF (S <> AMP.NAME)
	{
	    AMP.NAME = S
	    SEND_STRING VD_AMP, "'NAME:', AMP.NAME"
	}
	IF (AMP.STATUS = lsLOOKING)
	{
	    AMP.STATUS = lsFOUND	 	// ONLINE
	    SEND_STRING VD_AMP, 'ONLINE:1'
	    [VD_AMP, 1] = 0
	    ADD (acSYNC, 100)			// GET CHANNEL NAMES
	    ADD (acSYNC, 101)			// GET PRESET NAMES
	    ADD (acSYNC, 102)			// GET LEVELS
	}
    } 
    ELSE IF (RX_CMD = 'NACK' && LEN = 4)
    {
	RX_ERR_CODE = RX_DATA[1] * $1000000 +
		      RX_DATA[2] * $10000 +
		      RX_DATA[3] * $100 +
		      RX_DATA[4]
	SEND_STRING VD_AMP, "'ERR:', ITOHEX (RX_ERR_CODE)"
    }
    ELSE IF (RX_CMD = 'SYNC' && LEN > 4)
    { 
	Y = RX_DATA[1]
	IF (AMP.FAULTS <> RX_DATA[2])
	{
	    AMP.FAULTS = RX_DATA[2]
	    SEND_STRING VD_AMP, "'FAULTS:', ITOA (AMP.FAULTS)"
	}
	IF (AMP.OVERRIDEN <> RX_DATA[3])
	{
	    AMP.OVERRIDEN = RX_DATA[3]
	    SEND_STRING VD_AMP, "'OVERRIDE:', ITOA (AMP.OVERRIDEN)"
	}
	IF (AMP.POWERED = RX_DATA[4])
	{
	    AMP.POWERED = (!RX_DATA[4])
	    SEND_STRING VD_AMP, "'POWER:', ITOA (AMP.POWERED)"
	}
	IF (Y = 100 && LEN = 255)
	{
	    FOR (I = 0; I <= 3; I ++)
	    {
		S = MID_STRING (RX_DATA, 5 + (I * 31), 31)
		X = FIND_STRING (S, "$00", 1)
		IF (X) S = LEFT_STRING (S, X - 1)
		IF (S <> AMP.INPUT_NAME[I + 1])
		{
		    AMP.INPUT_NAME[I + 1] = S
		    SEND_STRING VD_AMP, "'INPUT_NAME:', ITOA (I + 1), ':', AMP.INPUT_NAME[I + 1]"
		}
	    }
	    FOR (I = 0; I <= 3; I ++)
	    {
		S = MID_STRING (RX_DATA, 129 + (I * 31), 31)
		X = FIND_STRING (S, "$00", 1)
		IF (X) S = LEFT_STRING (S, X - 1)
		IF (S <> AMP.OUTPUT_NAME[I + 1])
		{
		    AMP.OUTPUT_NAME[I + 1] = S
		    SEND_STRING VD_AMP, "'OUTPUT_NAME:', ITOA (I + 1), ':', AMP.OUTPUT_NAME[I + 1]"
		}
	    }
	    AMP.FORCED_STANDBY = RX_DATA[253]
	    AMP.FORCED_STANDBY_ALLOWED = RX_DATA[254]
	    IF (AMP.GLOBAL_MUTED <> RX_DATA[255])
	    {
		AMP.GLOBAL_MUTED = RX_DATA[255]
		SEND_STRING VD_AMP, "'MUTE:', ITOA (AMP.GLOBAL_MUTED)"
	    }
	}
	ELSE IF (Y = 101 && LEN = 164)
	{
	    FOR (I = 0; I <= 4; I ++)
	    {
		S = MID_STRING (RX_DATA, 5 + (I * 31), 31)
		X = FIND_STRING (S, "$00", 1)
		IF (X) S = LEFT_STRING (S, X - 1)
		AMP.PRESET_NAME[I + 1] = S
		AMP.PRESET_VALID[I + 1] = RX_DATA[160 + I]
	    }
	    UPDATE_PRESETS ()
	}
	ELSE IF (Y = 102 && LEN = 64)
	{
	    FOR (I = 0; I <= 3; I ++)
	    {
		X = 5 + (I * 15)
		L = RX_DATA[X + 13]
		M = RX_DATA[X + 14]
		IF (L <> AMP.LEVELS[I + 1] || M <> AMP.MUTED[I + 1])
		{
		    AMP.LEVELS[I + 1] = L
		    AMP.MUTED[I + 1] = M
		    SEND_STRING VD_AMP, "'OUTPUT:', ITOA (I + 1), ':', ITOA (L), ':', ITOA (M)"
		}
		FOR (J = 0; J <= 3; J ++)
		{
		    Y = (I * 4) + J + 1
		    L = RX_DATA[X + (J * 2) + 1]
		    M = RX_DATA[X + (J * 2) + 2]
		    IF (L > AMP.MIX[Y] || M <> AMP.MIXMUTED[Y])
		    {
			AMP.MIX[Y] = L
			AMP.MIXMUTED[Y] = M
			SEND_STRING VD_AMP, "'MIX:', ITOA (I + 1), ':', ITOA (J + 1), ':', ITOA (L), ':', ITOA (M)"
		    }
		}
	    }
	}
    }
}

DEFINE_FUNCTION CHAR[32] STATUS_NAME (INTEGER STATUS_)
{
    SWITCH (STATUS_)
    {
	CASE lsOFFLINE : RETURN 'OFFLINE'
	CASE lsLOOKING : RETURN 'LOOKING'
	CASE lsFOUND   : RETURN 'FOUND'
	DEFAULT        : RETURN "'UNKNOWN', ITOA (STATUS_)"
    }
}

DEFINE_FUNCTION INIT ()
{
    INTEGER I, J

    FOR (I = 1; I <= QUEUE_SIZE; I ++)
    {
	QUEUE.ID[I] = 0
	QUEUE.PARAM[I] = 0
    }
    QUEUE.COUNT = 0
    QUEUE.FULL = 0
    QUEUE.EMPTY = 1

    AMP.NAME = ''
    AMP.POWERED = 0
    AMP.GLOBAL_MUTED = 0
    AMP.FAULTS = 0
    AMP.STATUS = lsOFFLINE
    AMP.POLL = 0
    AMP.TIMEOUT = 0
    
    FOR (I = 1; I <= 16; I ++)
    {
	AMP.MIX[I] = 0
	AMP.MIXMUTED[I] = 0
    }
    FOR (I = 1; I <= 4; I ++)
    {
	AMP.LEVELS[I] = 0;
	AMP.MUTED[I] = 0;
	AMP.INPUT_NAME[I] = ''
	AMP.OUTPUT_NAME[I] = ''
    }
    FOR (I = 1; I <= 5; I ++)
    {
	AMP.PRESET_NAME[I] = ''
	AMP.PRESET_VALID[I] = 0
    }
    NO_ENTRIES = 0
    FOR (I = 1; I <= MAX_ENTRIES; I ++)
	ENTRY[I] = ''
}

(***********************************************************)
(*                STARTUP CODE GOES BELOW                  *)
(***********************************************************)
DEFINE_START

INIT ()

(***********************************************************)
(*                THE EVENTS GO BELOW                      *)
(***********************************************************)
DEFINE_EVENT

(* VIRTUAL DEVICE *)
DATA_EVENT [VD_AMP]
{	// COMMANDS TO AMP
    COMMAND :
    {
	INTEGER I, X, Y, Z
	
//	TERMINAL (DATA.TEXT)
	NO_ENTRIES = 0
	X = FIND_STRING (DATA.TEXT, ':', 1)
	WHILE (X)
	{
	    NO_ENTRIES ++
	    ENTRY[NO_ENTRIES] = LEFT_STRING (DATA.TEXT, X - 1)
	    GET_BUFFER_STRING (DATA.TEXT, X)
	    X = FIND_STRING (DATA.TEXT, ':', 1)
	}
	NO_ENTRIES ++
	ENTRY[NO_ENTRIES] = DATA.TEXT
	FOR (X = 1; X <= NO_ENTRIES; X ++)
	    TERMINAL ("'  ', ITOA (X), '  "', ENTRY[X], '"'")
	
	IF (NO_ENTRIES = 1)
	{
	    SWITCH (ENTRY[1])
	    {
		CASE 'REFRESH' :
		{
		    UPDATE_NAMES ()
		    UPDATE_PRESETS ()
		    UPDATE_STATUS ()
		    UPDATE_LEVELS ()
		}
	    }
	}
	ELSE IF (NO_ENTRIES = 2)
	{
	    SWITCH (ENTRY[1])
	    {
		CASE 'POWER' :
		{
		    SWITCH (ENTRY[2])
		    {
			CASE '0' : AMP.POWERED = 0
			CASE '1' : AMP.POWERED = 1
			CASE 'T' : AMP.POWERED = !AMP.POWERED
		    }
		    ADD (acPOWER, !AMP.POWERED)
		    SEND_STRING VD_AMP, "'POWER:', ITOA (AMP.POWERED)"
		}
		CASE 'MUTE' :
		{
		    SWITCH (ENTRY[2])
		    {
			CASE '0' : AMP.GLOBAL_MUTED = 0
			CASE '1' : AMP.GLOBAL_MUTED = 1
			CASE 'T' : AMP.GLOBAL_MUTED = !AMP.GLOBAL_MUTED
		    }
		    ADD (acMUTE, AMP.GLOBAL_MUTED)
		    SEND_STRING VD_AMP, "'MUTE:', ITOA (AMP.GLOBAL_MUTED)"
		}
		CASE 'PING_TIME' : PING_TIME = ATOI (ENTRY[2])
		CASE 'POLL_TIME' : POLL_TIME = ATOI (ENTRY[2])
		CASE 'PRESET' : 
		{
		    X = ATOI (ENTRY[2])
		    ADD (acPRESET, X)
		}
	    }
	}
	ELSE IF (NO_ENTRIES = 3)
	{
	    SWITCH (ENTRY[1])
	    {
		CASE 'OUTPUT' :
		{
		    X = ATOI (ENTRY[2])
		    IF (X && X <= 4)
		    {
			AMP.LEVELS[X] = ATOI (ENTRY[3])
			Z = acCH1 + ((X - 1) * 10) 
			TERMINAL ("'PRESENT ', ITOA (PRESENT (Z, 0))")
			IF (PRESENT (Z, 0) < 2) ADD (Z, 0)
			SEND_STRING VD_AMP, "'OUTPUT:', ENTRY[2], ':', ITOA (AMP.LEVELS[X]), ':', ITOA (AMP.MUTED[X])"
			IF (!QUEUE.COUNT && !QUEUE.EMPTY) EXEC ()
		    }
		}
		CASE 'OUTPUT_MUTE' :
		{
		    X  = ATOI (ENTRY[2])
		    IF (X && X <= 4)
		    {
			SWITCH (ENTRY[3])
			{
			    CASE '0' : AMP.MUTED[X] = 0
			    CASE '1' : AMP.MUTED[X] = 1
			    CASE 'T' : AMP.MUTED[X] = !AMP.MUTED[X]
			}
			Z = acCH1 + ((X - 1) * 10) 
			IF (PRESENT (Z, 0) < 2) ADD (Z, 0)
			SEND_STRING VD_AMP, "'OUTPUT:', ENTRY[2], ':', ITOA (AMP.LEVELS[X]), ':', ITOA (AMP.MUTED[X])"
			IF (!QUEUE.COUNT && !QUEUE.EMPTY) EXEC ()
		    }
		}
	    }
	}
	ELSE IF (NO_ENTRIES = 4)
	{
	    SWITCH (ENTRY[1])
	    {
		CASE 'MIX' :
		{
		    X = ATOI (ENTRY[2])
		    Y = ATOI (ENTRY[3])
		    IF (X && X <= 4 && Y && Y <= 4)
		    {
			I = (X - 1) * 4 + Y
			Z = acCH1IP1 + (Y - 1) + ((X - 1) * 10)  // COMMAND ID
			AMP.MIX[I] = ATOI (ENTRY[4])
			IF (PRESENT (Z, 0) < 2) ADD (Z, 0)
			SEND_STRING VD_AMP, "'MIX:', ENTRY[2], ':', ENTRY[3], ':', ITOA (AMP.MIX[I]), ':', ITOA (AMP.MIXMUTED[I])"
			IF (!QUEUE.COUNT && !QUEUE.EMPTY) EXEC ()
		    }
		}
		CASE 'MIX_MUTE' :
		{
		    X = ATOI (ENTRY[2])
		    Y = ATOI (ENTRY[3])
		    IF (X && X <= 4 && Y && Y <= 4)
		    {
			I = (X - 1) * 4 + Y
			Z = acCH1IP1 + (Y - 1) + ((X - 1) * 10)  // COMMAND ID
			SWITCH (ENTRY[4])
			{
			    CASE '0' : AMP.MIXMUTED[I] = 0
			    CASE '1' : AMP.MIXMUTED[I] = 1
			    CASE 'T' : AMP.MIXMUTED[I] = !AMP.MIXMUTED[I]
			}
			IF (PRESENT (Z, 0) < 2) ADD (Z, 0)
			SEND_STRING VD_AMP, "'MIX:', ENTRY[2], ':', ENTRY[3], ':', ITOA (AMP.MIX[I]), ':', ITOA (AMP.MIXMUTED[I])"
			IF (!QUEUE.COUNT && !QUEUE.EMPTY) EXEC ()
		    }
		}
	    }
	}
    }
}

BUTTON_EVENT [VD_AMP, 0]
{	// DIGITAL COMMANDS TO AMP
    PUSH :
    {
	SWITCH (BUTTON.INPUT.CHANNEL)
	{
	    CASE 1 : ADD (acPING, 0)
	    
	    CASE 2 : ADD (acSYNC, 100);
	    CASE 3 : ADD (acSYNC, 101);
	    CASE 4 : ADD (acSYNC, 102);
	    CASE 5 : ADD (acSYNC, 2)
	    CASE 6 : ADD (acSYNC, 6)
	
	
	    CASE 31 : TERMINAL ("ITOA (AMP_NO), ' STATUS ', STATUS_NAME (AMP.STATUS)")
	    CASE 32 : SEND_COMMAND VD_COMMS, "AMP_NO, '@AMP', AMP_ADDR"
	
	}
    }
}

// COMMS VIRTUAL DEVICE
DATA_EVENT [VD_COMMS]
{
    ONLINE :	// REGISTER AMPLIFIER WITH COMMS AND COMMENCE POLLING
    {
	SEND_COMMAND VD_COMMS, "AMP_NO, '@AMP', AMP_ADDR"
	SEND_STRING VD_AMP, 'ONLINE:0'
	AMP.STATUS = lsLOOKING
	AMP.POLL = RANDOM_NUMBER (30) + 100
	TIMELINE_CREATE (TL_TICK, TL_TICK_COUNT, 1, TIMELINE_ABSOLUTE, TIMELINE_REPEAT)
    }
    STRING : 	// RESPONSE FROM DEVICE
    {
	IF (LENGTH_STRING (DATA.TEXT) >= 5)
	{
	    RX_NO = GET_BUFFER_CHAR (DATA.TEXT)
	    IF (RX_NO = AMP_NO)
	    {
		RX_CMD = GET_BUFFER_STRING (DATA.TEXT, 4)
		RX_DATA = DATA.TEXT
		DECODE_DATA ()
	    }
	}
    }
}

TIMELINE_EVENT [TL_TICK]
{
    IF (QUEUE.COUNT) QUEUE.COUNT --
    IF (!QUEUE.COUNT && !QUEUE.EMPTY) EXEC ()
    IF (AMP.POLL)
    {
	AMP.POLL --
	IF (!AMP.POLL)
	{
	    SWITCH (AMP.STATUS)
	    {
		CASE lsLOOKING :
		{
		    TERMINAL ("'PINGING AMP ', ITOA (AMP_NO)")
		    ADD (acPING, 0)
		    AMP.POLL = PING_TIME
		}
		CASE lsFOUND :
		{
		    ADD (acSYNC, 102)
		    AMP.POLL = POLL_TIME
		}
	    }
	}
    }
    IF (AMP.TIMEOUT)
    {
	AMP.TIMEOUT --
	IF (!AMP.TIMEOUT)
	{
	    IF (AMP.STATUS = lsFOUND)
	    {
	  	SEND_STRING VD_AMP, 'ONLINE:0'
		[VD_AMP, 1] = 0
   		AMP.STATUS = lsLOOKING
		AMP.POLL = PING_TIME
	    } 
	}
    }  
}

(***********************************************************)
(*            THE ACTUAL PROGRAM GOES BELOW                *)
(***********************************************************)
DEFINE_PROGRAM

IF (!QUEUE.COUNT && !QUEUE.EMPTY) EXEC ()

(***********************************************************)
(*                     END OF PROGRAM                      *)
(*        DO NOT PUT ANY CODE BELOW THIS COMMENT           *)
(***********************************************************)
